/*
 * $Id: AmRtpPacket.cpp,v 1.3.2.1 2005/08/25 07:54:04 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmRtpPacket.h"
#include "rtp/rtp.h"
#include "log.h"

#include <assert.h>
#include <string.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

AmRtpPacket::AmRtpPacket()
    : b_size(0), data(0)
{
}

AmRtpPacket::~AmRtpPacket()
{
}

int AmRtpPacket::parse()
{
    assert(b_size);

    if (b_size<=sizeof(rtp_hdr_t)){
	ERROR("bad rtp packet (not enough space for the rtp header)!\n");
	return -1;
    }

    rtp_hdr_t* hdr = (rtp_hdr_t*)buffer;

    if(hdr->version != RTP_VERSION){
	DBG("received RTP packet with unsupported version (%i).\n",
	    hdr->version);
	return -1;
    }

    if(hdr->x != 0){
	DBG("RTP extension headers not supported.\n");
	return -1;
    }

    payload = hdr->pt;
    marker = hdr->m;
    sequence = ntohs(hdr->seq);
    timestamp = ntohl(hdr->ts);
    ssrc = ntohl(hdr->ssrc);

    data = buffer + sizeof(rtp_hdr_t) + (hdr->cc*4);
    if ((data - buffer)>=(long)b_size){
	ERROR("bad rtp packet (header size too big) !\n");
	return -1;
    }
    d_size = b_size - (data - buffer);

    if(hdr->p){
	if (data[d_size-1]>=d_size){
		ERROR("bad rtp packet (invalid padding size) !\n");
		return -1;
	}
	d_size -= data[d_size-1];
    }

    return 0;
}

int AmRtpPacket::compile(unsigned char* data_buf, unsigned int size)
{
    assert(data_buf);
    assert(size);

    d_size = size;
    b_size = d_size + sizeof(rtp_hdr_t);
    rtp_hdr_t* hdr = (rtp_hdr_t*)buffer;

    if(b_size>sizeof(buffer)){
	ERROR("builtin buffer size (%d) exceeded: %d\n",
		(int)sizeof(buffer), b_size);
	return -1;
    }

    memset(hdr,0,sizeof(rtp_hdr_t));
    hdr->version = RTP_VERSION;
    hdr->m = marker;
    hdr->pt = payload;
    
    hdr->seq = htons(sequence);
    hdr->ts = htonl(timestamp);
    hdr->ssrc = htonl(ssrc);
    
    data = buffer + sizeof(rtp_hdr_t);
    memcpy(data,data_buf,d_size);

    return 0;
}

int AmRtpPacket::send(int sd, const struct sockaddr_in* saddr)
{
    if(sendto(sd,buffer,b_size,0,
	      (const struct sockaddr *)saddr,
	      sizeof(struct sockaddr_in)) == -1){
	ERROR("while sending RTP packet: %s\n",strerror(errno));
	return -1;
    }

    return 0;
}

int AmRtpPacket::recv(int sd, struct sockaddr_in* recv_addr)
{
    socklen_t recv_addr_len = sizeof(struct sockaddr_in);

    int ret = recvfrom(sd,buffer,sizeof(buffer),0,
		       (struct sockaddr*)recv_addr,
		       &recv_addr_len);

    if(ret > 0){
	b_size = ret;
    }
    
    return ret;
}
