/*
 * $Id: IsdnCallManager.cpp,v 1.1.2.2 2004/01/23 14:05:32 ullstar Exp $
 *
 * Copyright (C) 2003 Ulrich Abend
 *
 * This file is part of isdngw, a free isdn gateway plugin 
 * for sems, a free SIP media server.
 *
 * isdngw is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * isdngw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "log.h"
#include "AmSession.h"

#include "globals.h"
#include "IsdnCallManager.h"
#include "IsdnDevice.h"
#include "IsdnConfiguration.h"
#include "IsdnDeviceIncoming.h"
#include "IsdnDevice.h"
#include "IsdnChat.h"
#include "IsdnUtils.h"
#include "IsdnCall.h"

/* globally available callManager instance */
IsdnCallManager* callManager = NULL;

IsdnCallManager::IsdnCallManager() {
	m_callDetails = new CallDetail();
}

IsdnCallManager::~IsdnCallManager() {
	delete m_callDetails;
}

void IsdnCallManager::instantiate() {
	if (callManager == NULL) {
		callManager = new IsdnCallManager();
		callManager->init();
	}
}

void IsdnCallManager::init() {
}

IsdnCall* IsdnCallManager::placeCallFromPSTN(char* fromnumber, char* tonumber, int backend, void* device) {
	IsdnCall* call = new IsdnCall();
	// TODO: try block to free call if unsuccessful
	call->placeCallFromPSTN(fromnumber, tonumber, backend, device);
	return call;
}

IsdnCall* IsdnCallManager::placeCallFromSIP(AmCmd cmd) {
	IsdnCall* call = new IsdnCall();
	// TODO: try block to free call if unsuccessful
	call->placeCallFromSIP(cmd);
	return call;
}

IsdnCall* IsdnCallManager::changeSIPState(AmCmd cmd, int newstate) {
	IsdnCall* call;
	// find right isdncall, call changeSIPState there
	// this is only called for oninvite for inbound call
	IsdnDeviceIncoming *matching_device = IsdnDeviceIncoming::getPendingCallDevice(cmd.user.c_str());
	if (matching_device != NULL) {
		/* found matching i4l device! */
		call = matching_device->getCall();
	} else {
		#ifdef USE_CAPI20
		IsdnCapiConnection* in_ccon = capiManager->getPendingCallConnection(cmd.user.c_str());
		if (in_ccon != NULL) {
			/* found matching cconnection */
			call = in_ccon->getCall();
		}	else {
		#endif
		ERROR("isdngw: internal error, no matching inbound call found\n");
		throw AmSession::Exception(500,"isdngw: internal error, no matching inbound call found\n");
		#ifdef USE_CAPI20
		}
		#endif
	}
	call->changeSIPState(newstate);
	call->setCommand(cmd);
	return call;
}

/* obsolete: void IsdnCallManager::changeSIPState(AmRequest *req, int newstate) {

}
*/

CallDetail* IsdnCallManager::addCall(IsdnCall* call) {
	DBG("new Call@%p for reporting app registered\n", call);
	/* create new callDetail object and attach to list */

	CallDetail* newdetail = new CallDetail();

	newdetail->next = NULL;
	newdetail->history = NULL;

	/*  not insert in list right now */
	/* 	CallDetail* cur = m_callDetails;
			while (cur->next != NULL)
				cur = cur->next;
			cur->next = newdetail;
	*/
	
	return newdetail;
}

void IsdnCallManager::diedCall(IsdnCall* call, CallDetail* callDetail) {
	
	// first fill a fields with valid information
	callDetail->from = (char*) malloc (strlen(call->getFromAsText())+1);
	if (callDetail->from == NULL) 
		ERROR("no memory for calldetail->from\n");
	else	
		strcpy(callDetail->from, call->getFromAsText());
	callDetail->to = (char*) malloc (strlen(call->getToAsText())+1);
	if (callDetail->to == NULL) 
		ERROR("no memory for calldetail->to\n");
	else	
		strcpy(callDetail->to, call->getToAsText());
	
	callDetail->device = (char*) malloc (strlen(call->getDeviceAsText())+1);
	if (callDetail->device == NULL) 
		ERROR("no memory for calldetail->device\n");
	else	
		strcpy(callDetail->device, call->getDeviceAsText());
	
	callDetail->backend = call->getBackendAsText();
	
	DBG("Call object deleted, dumping details of call\n");
	dumpDetails(callDetail);
}

void IsdnCallManager::updateCall(IsdnCall* call, CallDetail* callDetail, char* entry) {
	/* report a status change */
	History* newhistory = new History();
	gettimeofday(&newhistory->timestamp, NULL);
	newhistory->description = (char*) malloc(strlen(entry)+1);
	if (newhistory->description == NULL) 
		ERROR("no memory to store call detail history entry\n");
	sprintf(newhistory->description, "%s", entry); 
	newhistory->next = NULL;
	History* cur = callDetail->history;
	if (cur == NULL) { /* no history yet */
		callDetail->history = newhistory;
	} else { /* append */
		while (cur->next != NULL)
			cur = cur->next;
		cur->next = newhistory;
	}	
}

void IsdnCallManager::dumpDetailsOfCalls() {
	int i = 0;
	for (CallDetail* cur = m_callDetails; cur!=NULL; cur = cur->next, i++) {
		dumpDetails(cur);
	}	
}

/* recursively free history */
void freeHistory (History* hist) {
	if (hist->next != NULL)
		freeHistory(hist->next);
	if (hist == NULL)
		return;
	free(hist->description);
	delete hist;
}	

void IsdnCallManager::dumpDetails(CallDetail* calldetail){
	History* cur = calldetail->history;
	DBG("dumping details for detail record@%p\n", calldetail);

	DBG("Call from %s to %s using device %s from backend %s\n", calldetail->from, calldetail->to, calldetail->device, calldetail->backend);
	if (calldetail->history == NULL) {
		DBG("  -> NO HISTORY\n");
	} else {
		do {
			/* DBG("  -> %u: %s (PSTN connected? %s)\n", cur->timestamp.tv_sec, cur->description, cur->pstnConnected?"yes":"no"); */
			DBG("  -> %u: %s", (unsigned) cur->timestamp.tv_sec, cur->description);
			cur = cur->next;
		} while (cur != NULL);
			
	}
	freeHistory(calldetail->history);
	delete calldetail;
}
