/*
 * $Id: AnswerMachine.cpp,v 1.12.2.4 2004/11/11 12:22:54 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * For a license to use the sems software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AnswerMachine.h"
#include "AmApi.h"
#include "AmSession.h"
#include "AmConfig.h"
#include "AmMail.h"
#include "AmUtils.h"

#include "sems.h"
#include "log.h"

#include <unistd.h>

#define DEFAULT_MAIL_TMPL   "/usr/local/lib/sems/plug-in/mail.template"
#define MOD_NAME "voicemail"

EXPORT_FACTORY(AnswerMachineFactory,MOD_NAME);

AnswerMachineFactory::AnswerMachineFactory(const string& _app_name)
  : AmStateFactory(_app_name)
{}

string AnswerMachineFactory::AnnouncePath;
string AnswerMachineFactory::DefaultAnnounce;
int    AnswerMachineFactory::MaxRecordTime;

int AnswerMachineFactory::onLoad()
{
    char* p = 0;

    if(m_config.reloadModuleConfig(MOD_NAME) != 1)
	return -1;

    if( ((p = m_config.getValueForKey("announce_path")) != NULL) && (*p != '\0') )
	AnnouncePath = p;
    else {
	WARN("no announce_path specified in configuration\n");
	WARN("file for module voicemail: using default.\n");
	AnnouncePath = ANNOUNCE_PATH;
    }

    if( ((p = m_config.getValueForKey("default_announce")) != NULL) && (*p != '\0') )
	DefaultAnnounce = p;
    else {
	WARN("no default_announce specified in configuration\n");
	WARN("file for module voicemail: using default.\n");
	DefaultAnnounce = DEFAULT_ANNOUNCE;
    }

    if( ((p = m_config.getValueForKey("max_record_time")) != NULL) && (*p != '\0') ){
 	if(sscanf(p,"%u",&MaxRecordTime) != 1) {
	    ERROR("could not convert max_record_time ('%s') to an integer.\n",p);
	    ERROR("using default max_record_time.\n");
	    MaxRecordTime = DEFAULT_RECORD_TIME;
 	}
    }
    else {
	WARN("no max_record_time specified in configuration\n");
	WARN("file for module voicemail: using default.\n");
	MaxRecordTime = DEFAULT_RECORD_TIME;
    }

    string email_tmpl_file;
    if( ((p = m_config.getValueForKey("email_template")) != NULL) && (*p != '\0') ){
	email_tmpl_file = p;
    }
    else {
	WARN("no email_template specified in configuration\n");
	WARN("file for module voicemail: using default.\n");
	email_tmpl_file = DEFAULT_MAIL_TMPL;
    }
    
    if(email_tmpl.load(email_tmpl_file)){
	ERROR("Voicemail: could not load email template: '%s'\n",email_tmpl_file.c_str());
	return -1;
    }
    
    return 0;
}

AmDialogState* AnswerMachineFactory::onInvite(AmCmd& cmd)
{
    string announce_path = AnnouncePath;
    if( !announce_path.empty() 
	&& announce_path[announce_path.length()-1] != '/' )
	announce_path += "/";
    
    string announce_file = announce_path + cmd.domain 
	+ "/" + cmd.user + ".wav";

    if(!file_exists(announce_file)){
	announce_file = announce_path + cmd.domain 
	    + "/" + DefaultAnnounce;

	if(!file_exists(announce_file)){
	    announce_file = announce_path + DefaultAnnounce;
	    if(!file_exists(announce_file))
		announce_file = "";
	}
    }


    if(announce_file.empty())
	throw AmSession::Exception(500,"voicemail: no greeting file found");
    
    if(cmd.email.empty())
	throw AmSession::Exception(404,"voicemail: no email address for user <" 
				   + cmd.user + ">");
    
    DBG("email address for user '%s': %s\n",
	cmd.user.c_str(),cmd.email.c_str());
    
    return new AnswerMachineDialog(announce_file,&email_tmpl);
}

AnswerMachineDialog::AnswerMachineDialog(string announce_file, const EmailTemplate* tmpl) 
    : announce_file(announce_file), tmpl(tmpl)
{
}

AnswerMachineDialog::~AnswerMachineDialog()
{
}

void AnswerMachineDialog::onSessionStart(AmRequest* req)
{
    DBG("starting sending announce...(%s)\n",announce_file.c_str());
    getSession()->rtp_str.play(announce_file);
    getSession()->rtp_str.play(AnswerMachineFactory::AnnouncePath + "/beep.wav");
    DBG("... announce reaches its end.\n");

    if(!getSession()->sess_stopped.get()){
	
	string msg_filename = string("/tmp/") + req->cmd.callid + string(".wav");
	    
	DBG("starting recording message...(%s)\n",msg_filename.c_str());
	float rec_length=0.0;
	int rec_size = getSession()->rtp_str.record( msg_filename.c_str(),
						     AnswerMachineFactory::MaxRecordTime,
						     rec_length);
	DBG("... finished recording message\n");
	    
	    
	if(!getSession()->sess_stopped.get())
	    req->bye();
	    
	DBG("recorded data size: %i\n",rec_size);

	if(rec_size){
	    try {
#ifdef _OBSOLETE
		string body = "Hello " + req->cmd.to + ",\n";
		body += req->cmd.from + " left a voice message for you.\n\n";
		body += "Thank you for using iptel.org's Voicemail.\n\n";
		body += "Your voicemail system.\n\n";
		body += "---------------------------------------------------------\n";
		body += "This message was created by iptel.org's voicemail system.\n";
		body += "Please review www.iptel.org for more information.\n";
		body += "---------------------------------------------------------\n";
		
		AmMail* mail = new AmMail("voicemail@iptel.org",
					  "Voice message from: " + req->cmd.from,
					  req->cmd.email,body);
#else
		AmMail* mail = new AmMail(tmpl->getEmail(req->cmd));
#endif
		
		mail->attachements.push_back(Attachement(msg_filename,
							 "message.wav",
							 "audio/x-wav"));
		mail->clean_up = clean_up_mail;
		AmMailDeamon::instance()->sendQueued(mail);
	    }
	    catch(const string& err){
		ERROR("while creating email: %s\n",err.c_str());
	    }
	}
    }
}

void AnswerMachineDialog::onBye(AmRequest* req)
{
}

void AnswerMachineDialog::clean_up_mail(AmMail* mail)
{
    for( Attachements::const_iterator att_it = mail->attachements.begin();
	 att_it != mail->attachements.end(); ++att_it )
	
	unlink(att_it->fullname.c_str());
}
